"""This file has functions that handle the user cli commands"""

import subprocess
import shutil
from src.pkg import utils


def getComposeConfig(username, server, path):
    """Makes and returns the config for the user"""

    template = {}
    mapping = {
        "${DTAAS_DIR}": path,
        "${username}": username,
    }
    try:
        if server == utils.LOCALHOST_SERVER:
            template, err = utils.importYaml("users.local.yml")
            utils.checkError(err)

        else:
            template, err = utils.importYaml("users.server.yml")
            utils.checkError(err)
            mapping["${SERVER_DNS}"] = server

        config, err = utils.replaceAll(template, mapping)
        utils.checkError(err)
    except Exception as e:
        return None, e

    return config, None


def createUserFiles(users, filePath):
    """Creates all the users' workspace directories"""
    for username in users:
        shutil.copytree(
            filePath + "/template", filePath + "/" + username, dirs_exist_ok=True
        )


def addUsersToCompose(users, compose, server, path):
    """Adds all the users config to the compose dictionary"""
    for username in users:
        config, err = getComposeConfig(username, server, path)
        if err is not None:
            return err
        compose["services"][username] = config
    return None


def startUserContainers(users):
    """Starts all the user containers in the 'users' list"""

    cmd = "docker compose -f compose.users.yml up -d"
    err = runCommandForContainers(cmd, users)
    return err


def stopUserContainers(users):
    """Stops all the user containers in the 'users' list"""

    cmd = "docker compose -f 'compose.users.yml' down"
    err = runCommandForContainers(cmd, users)
    return err


def runCommandForContainers(command, containers):
    cmd = [command]
    for name in containers:
        cmd.append(name)

    cmdStr = " ".join(cmd)
    result = subprocess.run(cmdStr, shell=True, check=False)
    if result.returncode != 0:
        return Exception(f"failed to run '{cmdStr}' command")
    return None


def addUsers(configObj):
    """add cli command handler"""
    try:
        compose, err = utils.importYaml("compose.users.yml")
        utils.checkError(err)
        userList, err = configObj.getAddUsersList()
        utils.checkError(err)
        server, err = configObj.getServerDNS()
        utils.checkError(err)
        path, err = configObj.getPath()
        utils.checkError(err)
    except Exception as e:
        return e

    if "version" not in compose:
        compose["version"] = "3"
    if "services" not in compose:
        compose["services"] = {}
    if "networks" not in compose:
        compose["networks"] = {"users": {"name": "dtaas-users", "external": True}}

    try:
        err = createUserFiles(userList, path + "/files")
        utils.checkError(err)
        err = addUsersToCompose(userList, compose, server, path)
        utils.checkError(err)
        err = utils.exportYaml(compose, "compose.users.yml")
        utils.checkError(err)
        err = startUserContainers(userList)
        utils.checkError(err)
    except Exception as e:
        return e

    return None


def deleteUser(configObj):
    """delete cli command handler"""
    try:
        compose, err = utils.importYaml("compose.users.yml")
        utils.checkError(err)
        userList, err = configObj.getDeleteUsersList()
        utils.checkError(err)
        err = stopUserContainers(userList)
        utils.checkError(err)

        for username in userList:
            if "services" in compose and username in compose["services"]:
                del compose["services"][username]

        err = utils.exportYaml(compose, "compose.users.yml")
        utils.checkError(err)

    except Exception as e:
        return e

    return None



"This file has generic helper functions and variables for dtaas cli"

import yaml
import tomlkit

LOCALHOST_SERVER = "localhost"


def importYaml(filename):
    """This function is used to import a yaml file safely"""
    config = {}
    try:
        with open(filename, "r") as file:
            config = yaml.safe_load(file)
    except FileNotFoundError:
        return {}, None
    except Exception as err:
        return None, Exception(
            f"Error while getting yaml file: {filename}, " + str(err)
        )
    return config if config is not None else {}, None


def exportYaml(data, filename):
    """This function is used to export to a yaml file safely"""
    try:
        with open(filename, "w") as file:
            yaml.safe_dump(data, file, sort_keys=False)
    except Exception as err:
        return Exception(f"Error while writing yaml to file: {filename}, " + str(err))
    return None


def importToml(filename):
    """This function is used to import a toml file safely"""
    try:
        with open(filename, "r") as file:
            config = tomlkit.load(file)
    except Exception as err:
        return None, Exception(
            f"Error while getting toml file: {filename}, " + str(err)
        )
    return config, None


def replaceAll(obj, mapping):
    """This function is used to replace all placeholders with values in a nested object"""
    if isinstance(obj, str):
        obj, err = replaceString(obj, mapping)
        return obj, err

    if isinstance(obj, list):
        obj, err = replaceList(obj, mapping)
        return obj, err

    if isinstance(obj, dict):
        obj, err = replaceDict(obj, mapping)
        return obj, err

    return None, Exception("Config substition failed: Object format not valid")


def replaceString(s, mapping):
    for key in mapping:
        s = s.replace(key, mapping[key])
    return s, None


def replaceList(arr, mapping):
    for ind, val in enumerate(arr):
        arr[ind], err = replaceAll(val, mapping)
        if err is not None:
            return None, err
    return arr, None


def replaceDict(dictionary, mapping):
    for key in dictionary:
        if not isinstance(key, str):
            return None, Exception("Config substitution failed: Key is not a string")
        dictionary[key], err = replaceAll(dictionary[key], mapping)
        if err is not None:
            return None, err
    return dictionary, None


def checkError(err):
    if err is not None:
        raise err



"""This file supports the DTaaS config class"""

import click
from src.pkg import utils


class Config:
    """The Config class for DTaaS"""

    def __init__(self):
        config, err = utils.importToml("dtaas.toml")
        if err is not None:
            raise click.ClickException("config initialisation failed: " + str(err))
        self.data = config

    def getConfig(self):
        """Gets the config dictionary"""
        if self.data is not None:
            return self.data, None
        return None, Exception("Config not initialised")

    def getFromConfig(self, key):
        """Gets the specific key from config"""
        conf, err = self.getConfig()
        if err is not None:
            return None, err

        if key not in conf:
            return None, Exception(f"Config file error: Missing {key} tag")
        return conf[key], None

    def getCommon(self):
        """Gets the 'common' section of config"""
        common, err = self.getFromConfig("common")
        return common, err

    def getStringFromCommon(self, key):
        """Gets the specific string key from config.common"""
        confCommon, err = self.getCommon()
        if err is not None:
            return None, err

        if key not in confCommon or confCommon[key] == "":
            return None, Exception(
                f"Config file error: config.common.{key} not set in TOML"
            )
        return str(confCommon[key]), None

    def getUsers(self):
        """Gets the 'users' section of config"""
        users, err = self.getFromConfig("users")
        return users, err


    def getStringListFromUsers(self, key):
        """Gets the specific key as a list of strings from config.users"""
        err = None
        stringsList = None

        confUsers, err = self.getUsers()
        if err is not None:
            return None, err

        try:
            stringsList = [ str(x) for x in confUsers[key]]
            if len(stringsList) == 0:
                stringsList = None
                raise ValueError
        except KeyError:
            err = Exception(f"Config file error: No {key} list in 'users' tag")
        except ValueError:
            err = Exception(f'Config file error: users.{key} list is empty')

        return stringsList, err

    def getPath(self):
        """Gets the 'path' from config.common"""
        path, err = self.getStringFromCommon("path")
        return path, err

    def getServerDNS(self):
        """Gets the 'server-dns' from config.common"""
        server, err = self.getStringFromCommon("server-dns")
        return server, err

    def getAddUsersList(self):
        """Gets the 'add' list from config.users"""
        addUsersList, err = self.getStringListFromUsers("add")
        return addUsersList, err

    def getDeleteUsersList(self):
        """Gets the 'delete' list from config.users"""
        deleteUsersList, err = self.getStringListFromUsers("delete")
        return deleteUsersList, err
